// 0: anyone can use /ghostAllBricks
// 1: only admins can use it
// 2: only super admins can use it
if ($Pref::Server::GhostAllAdmin $= "")
  $Pref::Server::GhostAllAdmin = 1;

if ($Pref::Server::GhostAllTickCount $= "")
  $Pref::Server::GhostAllTickCount = 400;

if ($Pref::Server::GhostAllTickDelay $= "")
  $Pref::Server::GhostAllTickDelay = 20;

function serverCmdRestartGhosting(%client)
{
    // If they haven't finished loading yet, they haven't started ghosting yet.
    // This means that starting ghosting prematurely could ghost objects which
    // they do not have the datablocks for (which is, of course, bad).
    if (!%client.hasSpawnedOnce)
        return;

    // If they aren't admin, add some degree of flood protection.
    if (!%client.isAdmin)
    {
        %elapsed = $Sim::Time - %client.lastRestartGhostingTime;
        %remaining = 6 - %elapsed;

        if (%remaining > 0)
        {
            messageClient(%client, '', "\c2You cannot restarting ghosting again so soon. Please wait \c3" @ mCeil(%remaining) @ " second(s) \c2before trying again.");
            return;
        }

        %client.lastRestartGhostingTime = $Sim::Time;
    }

    messageClient(%client, '', "\c2Restarting the ghosting process.");

    if (getSubStr(getDateTime(), 0, 5) $= "05/07") // What could it be?
        messageClient(%client, '', "\c5This is going to feel... a little weird.");

    // Magic trick.
    %client.resetGhosting();
    %client.activateGhosting();
}

function serverCmdGhostAllBricks(%client)
{
  // Find the "permission level" (defined above GhostAllAdmin pref) of the user.
  %admin = %client.isSuperAdmin ? 2 : (%client.isAdmin ? 1 : 0);

  // Make sure they're at least at or above the required level before continuing.
  if (%admin < $Pref::Server::GhostAllAdmin)
  {
    messageClient(%client, '', "\c2You do not have permission to use this command.");
    return;
  }

	messageClient(%client, '', "\c2Sending all bricks from server.");
	%client.ghostAllTick(0, 0); // will cancel an existing loop
}

// In case the server has a lot of bricks, we don't want to loop over all of them at once.
// That could lock up the server for a second or two. Instead, we just send several bricks
// at a time with a repeating schedule.
function GameConnection::ghostAllTick(%this, %groupIndex, %brickIndex)
{
	cancel(%this.ghostAllTick);

  // In case this value is large, don't read the global every single iteration.
  %tickCount = $Pref::Server::GhostAllTickCount;

  // This isn't necessarily, but it's clearer. Why would I leave a hack like this in
  // when I'm going through all the effort to place these comments here for no reason?
  // Yes, that's you! I'm talking about you! What are you doing here reading these comments?!
  // Did you want to learn how add-ons work? Great! I hope you appreciate how much I'm explaining the code.
  // Did you want to check for exploits to use them? That isn't very nice of you.
  // Did you want to check for exploits to warn others? It isn't very nice of you to suspect that I would
  // put something like that in here, but it's good to be critical anyway. If there's an exploit that I
  // *accidentally* left in, consider telling me first before posting it so I can release a fix before
  // everybody goes on to abuse it.
  %brickCount = 0;

	while (%brickCount < %tickCount)
	{
		if (%groupIndex >= MainBrickGroup.getCount())
		{
			messageClient(%this, '', "\c2Done sending bricks from server, watch ghosting indicator.");
			return;
		}

		%group = MainBrickGroup.getObject(%groupIndex);

    // Have we run out of bricks in this group? We need to go to the next (or stop).
		if (%brickIndex >= %group.getCount())
		{
      // Just go into the next iteration of the same loop again,
      // since it deals with checking whether or not this was the last group,
      // etc. (everything we need here).
			%groupIndex++;
			%brickIndex = 0;
			continue;
		}

		%group.getObject(%brickIndex).scopeToClient(%this);

		%brickIndex++;
		%brickCount++; // for max bricks per tick
	}

  // Since the only state used by the loop is contained in these two variables,
  // just schedule function over and over, passing these locals unmodified.
  // The loop makes sure they have their correct incrementing values.
	%this.ghostAllTick = %this.schedule($Pref::Server::GhostAllTickDelay, "ghostAllTick", %groupIndex, %brickIndex);
}

// Hooks into default game code.
// There's a decent explanation of how packages work in a sticky in Modification Help.
// Go there if all of this seems confusing.
package GhostUtils
{
  // When you fire and hit something with the brick tool ...
  function BrickDeployProjectile::onCollision(%this, %obj, %col, %fade, %pos, %normal)
  {
    %player = %obj.sourceObject;
    %client = %obj.client;

    %prevTempBrick = %player.tempBrick;

    // How did I find these arguments? I'll admit it, I just searched for "onCollision" on
    // the forums. It's lazy, this is something I should just remember, etc. It's fine, though.
    // If you can't find any documentation on the arguments for a function that way,
    // look up `trace`, or try packaging it and just printing out the arguments.
    // Make some guesses!
    Parent::onCollision(%this, %obj, %col, %fade, %pos, %normal);

    %tempBrick = %player.tempBrick;

    // ... check if it has changed your ghost brick, and if it has ...
    if (isObject(%tempBrick) && %tempBrick !$= %prevTempBrick && isObject(%client))
    {
      // ... then make sure your client tries to ghost it immediately.
      %tempBrick.scopeToClient(%obj.client);
    }

    // We check whether or not the ghost brick has changed because there's no need to call
    // `scopeToClient` again otherwise. It just moves the existing ghost brick.
  }

  // The new duplicator uses `onPlant` and already scopes bricks itself,
  // so we're using this instead. It's a decent entry point, and doesn't
  // attempt to scope the new brick when it isn't necessary.
  function FxDTSBrick::trustCheckFinished(%this)
  {
    Parent::trustCheckFinished(%this);

    if (isObject(%this.client))
      %this.scopeToClient(%this.client);
  }
};

activatePackage("GhostUtils");
